"""CLI module.

This module serves as the entrypoint to the program.

* ``AliasedGroup`` enables command aliases.
* ``opx`` runs the program itself, and is used as the entrypoint.
"""
import sys

from pathlib import Path

import click
import click_completion
import click_log

from opx import workspace
from opx.cmd import (
    assemble, build, init, release, remove, shell, status,
)

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

click_completion.init()
click_log.basic_config('opx')


class AliasedGroup(click.Group):

    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx)
                   if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail('Too many matches: %s' % ', '.join(sorted(matches)))


@click.group(cls=AliasedGroup, context_settings=CONTEXT_SETTINGS)
@click.version_option(prog_name='OpenSwitch Development Tool')
@click_log.simple_verbosity_option('opx')
@click.pass_context
def opx(ctx):
    parser = workspace.parser_read()
    workspace.cleanup_deleted_workspaces(parser)
    workspace.cleanup_removed_containers(parser)
    workspace.parser_write(parser)

    pwd = Path.cwd()

    try:
        ctx.obj = workspace.Workspace.load(pwd)
    except workspace.ForeignWorkspaceException:
        click.secho('I have no knowledge of this workspace.',
                    fg='magenta')
        click.secho('Shall I add it to my archives? [yn] ',
                    fg='magenta', nl=False)
        c = click.getchar()
        click.echo()
        if c == 'y':
            ctx.obj = workspace.Workspace(pwd, temp=False)
            ctx.obj.save()
        elif c == 'n':
            ctx.obj = workspace.Workspace(pwd)
        else:
            click.secho('Invalid key. Aborting...', fg='red')
            sys.exit(1)
    except workspace.InvalidWorkspaceException:
        ctx.obj = workspace.Workspace(pwd)


opx.add_command(assemble)
opx.add_command(build)
opx.add_command(init)
opx.add_command(release)
opx.add_command(remove)
opx.add_command(shell)
opx.add_command(status)



from opx.cmd.assemble import assemble
from opx.cmd.build import build
from opx.cmd.init import init
from opx.cmd.remove import remove
from opx.cmd.release import release
from opx.cmd.shell import shell
from opx.cmd.status import status



import logging

import click
import docker

from beautifultable import BeautifulTable

from opx import workspace

L = logging.getLogger('opx')


@click.command()
def status() -> None:
    """Show overview of workspaces.

    Lists workspace directory, container image, and container name.
    """
    table = BeautifulTable()
    table.left_border_char = ''
    table.right_border_char = ''
    table.top_border_char = ''
    table.bottom_border_char = ''
    # yes, separator is spelled incorrectly in the library
    table.row_separator_char = ''
    table.header_separator_char = ''
    table.column_headers = ['workspace', 'dist', 'image', 'container']
    table.column_alignments['workspace'] = BeautifulTable.ALIGN_LEFT
    table.column_alignments['dist'] = BeautifulTable.ALIGN_LEFT
    table.column_alignments['image'] = BeautifulTable.ALIGN_LEFT
    table.column_alignments['container'] = BeautifulTable.ALIGN_LEFT

    parser = workspace.parser_read()
    client = docker.from_env()

    for root in parser.sections():
        row = [root, parser[root]['dist'], '', '']
        if parser[root].get('cid', '') != '':
            container = client.containers.get(parser[str(root)].get('cid'))

            # Check if running container is outdated.
            try:
                row[2] = container.image.tags[0]
            except IndexError:
                L.info('A new image is available.')
                L.info('You can update by removing the current container.')
                pass

            row[3] = container.name
        table.append_row(row)

    L.info(table)



import gzip
import logging

from pathlib import Path

import requests

from deb_pkg_tools.control import deb822_from_string, parse_control_fields

L = logging.getLogger('opx')

DUMB_BUILD_ORDER = [
    'opx-core',
    'opx-logging',
    'opx-common-utils',
    'opx-cps',
    'opx-base-model',
    'opx-db-sql',
    'opx-sai-api',
    'opx-sai-vm',
    'opx-nas-common',
    'opx-nas-linux',
    'opx-nas-ndi-api',
    'opx-nas-ndi',
    'opx-nas-acl',
    'opx-nas-interface',
    'opx-nas-l2',
    'opx-nas-l3',
    'opx-nas-qos',
    'opx-nas-daemon',
    'opx-platform-config',
    'opx-sdi-sys',
    'opx-pas',
    'opx-tmpctl',
    'opx-tools',
    'opx-alarm',
]


def remote_packages(url: str):
    """Returns a dict containing the latest version of each package.

    Args:
        url: Points to a Packages.gz file

    Returns:
        dict[str]dict: a unique collection of packages
    """
    packages_dict = {}

    response = requests.get(url)

    for pkg_str in gzip.decompress(response.content).decode().split('\n\n'):
        pkg = parse_control_fields(deb822_from_string(pkg_str))
        if pkg.get('Package') and pkg.get('Package') not in packages_dict:
            packages_dict[pkg.get('Package')] = pkg

    return packages_dict


def resolve_dependencies(repo: str, control, packages):
    """Transform repo Build-Depends into source packages.

    Args:
        repo: Name of package we want dependencies for
        control (dict): Returned from parse_control_fields
        packages (dict[str]dict): Dict of dicts returned from
           parse_control_fields, with the package name as the key

    Returns:
        list[str]: Source packages to build sorted by DUMB_BUILD_ORDER
    """
    # TODO: Recursively fetch build dependencies
    rv = []
    for d in control['Build-Depends'].names:
        if d in packages:
            rv.append(packages[d].get('Source'))

    # TODO: Properly sort build dependencies
    rv.sort(key=lambda x: DUMB_BUILD_ORDER.index(x))

    L.debug(f'Need to build in this order: {rv}')

    rv.append(repo)
    return rv


def opx_build_dependencies(repo: str):
    """Returns a list of opx build dependencies for a repository.

    Args:
        repo: Repository to create a build order for

    Returns:
        list[str]: An order of repositories to build
    """
    if repo == 'all':
        return DUMB_BUILD_ORDER

    if not Path(f'{repo}/debian/control').exists():
        return [repo]

    controlfile = Path(f'{repo}/debian/control')
    control = parse_control_fields(deb822_from_string(controlfile.read_text()))

    # TODO: Cache this file in config directory
    packages = remote_packages(
        'http://deb.openswitch.net/dists/unstable/opx/binary-amd64/Packages.gz',
    )

    return resolve_dependencies(repo, control, packages)



import logging
import sys

from pathlib import Path
from typing import Tuple

import click
import sh

from opx import workspace

L = logging.getLogger('opx')


@click.command()
@click.option('--dist', default='unstable',
              help='OPX distribution to build against.')
@click.option('--url', '-u',
              default='git://git.openswitch.net/opx/opx-manifest',
              help='Manifest repository URL')
@click.option('--manifest', '-m', default='default.xml',
              help='Manifest file within repository')
@click.option('--branch', '-b', default='master',
              help='Branch of repository to use')
@workspace.pass_context
def init(
        ws: workspace.Workspace,
        dist: str,
        url: str,
        manifest: str,
        branch: str
):
    """Initialize an OpenSwitch workspace.

    Runs repo init and repo sync. DIST will be saved in workspace configuration.

    URL, MANIFEST, BRANCH, and PROJECTS are passed directly to repo.

    Examples:

    opx init

        Create a workspace with the latest content from each repository.

    opx init --dist stable

        Create a workspace with the latest stable packages.

    opx init --dist 2.2.1

        Create a workspace for the 2.2.1 release.

    opx init --release 2.2 (coming soon)

        Create a workspace with each repository's HEAD pointing to the tag of
        the package in the release. When building, missing dependencies are
        pulled from the exact set of packages used to build the release
        initially. A release can be recreated from code by running:

            opx build all && opx assemble --dist 2.2 -n 0
    """
    if ws.temp:
        root = Path.cwd()

        L.info('Initializing repositories...')
        try:
            # When we want streaming stdout/err, use _fg=True
            sh.repo.init(u=url, m=manifest, b=branch, _fg=True)
        except sh.ErrorReturnCode as e:
            L.error(f'RAN: {e.full_cmd}')
            sys.exit(1)

        ws = workspace.Workspace(root, dist, temp=False)
        ws.save()

        L.info('Synchronizing repositories...')
        try:
            sh.repo.sync(_fg=True)
        except sh.ErrorReturnCode as e:
            L.error(f'RAN: {e.full_cmd}')
            sys.exit(1)
    else:
        L.error('This workspace has already been initialized.')



import logging

import click

from opx import workspace

L = logging.getLogger('opx')


@click.command()
@click.option('--blueprint', '-b',
              default='opx-onie-installer/release_bp/OPX_dell_base.xml',
              help='Installer blueprint to use')
@click.option('--number', '-n', default=9999,
              help='Release number')
@click.option('--suffix', '-s', default='',
              help='Release suffix')
@click.option('--debug/--no-debug', default=False,
              help='Run opx_rel_pkgasm with --debug.')
@click.option('--dist',
              help='OPX distribution to build against.')
@click.option('--rm/--no-rm', default=False,
              help='Remove container when command exits.')
@workspace.pass_context
def assemble(
        ws: workspace.Workspace,
        blueprint: str,
        number: int,
        suffix: str,
        debug: bool,
        dist: str,
        rm: bool
):
    """Assemble an OpenSwitch ONIE installer.

    Local packages always have the highest priority. If packages are not found
    locally, they are pulled from the chosen DIST on deb.openswitch.net.
    The DIST chosen is used in the image for package updates.

    Examples:

    opx assemble

        Create an ONIE installer. Installer will have version "unstable.9999".
        Packages will be pulled from 'http://deb.openswitch.net/ unstable main'

    opx assemble --dist testing -n 5

        Create an ONIE installer. Installer will have version "testing.5".
        Packages will be pulled from 'http://deb.openswitch.net/ testing main'

    opx assemble --dist 2.2.0 -n 0

        Create an ONIE installer. Installer will have version "2.2.0".
        Packages will be pulled from 'http://deb.openswitch.net/ 2.2.0 main'
   """
    dist = dist or ws.dist
    cmd = f'opx_rel_pkgasm.py -b {blueprint} -n {number} --dist {dist}'
    if suffix != "":
        cmd += f' -s {suffix}'
    if debug:
        cmd += ' --debug -v 9'

    ws.run(rm, dist or ws.dist, cmd)



import logging

import click

from opx import workspace

L = logging.getLogger('opx')


@click.command()
@click.option('--dist',
              help='Change OPX distribution to build against.')
@click.option('--rm/--no-rm', default=False,
              help='Remove container when command exits.')
@workspace.pass_context
def shell(ws: workspace.Workspace, dist: str, rm: bool) -> None:
    """Launch an OpenSwitch development container.

    Creates a build container and runs bash inside.

    Build dependencies are pulled from the DIST distribution of OPX if not
    present locally.
    """
    ws.run(dist=dist or ws.dist, remove=rm)
